/* 
 * Copyright (c) 2015 Codethink Ltd
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef IVI_INPUT_CLIENT_PROTOCOL_H
#define IVI_INPUT_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct ivi_input;

extern const struct wl_interface ivi_input_interface;

/**
 * ivi_input - controller interface to the input system
 * @seat_created: A new seat has appeared
 * @seat_capabilities: A seat's capabilities have changed
 * @seat_destroyed: A seat is no longer available
 * @input_focus: input focus of surface in ivi compositor has changed
 * @input_acceptance: input acceptance has changed
 *
 * This includes handling the existence of seats, seat capabilities, seat
 * acceptance and input focus.
 */
struct ivi_input_listener {
	/**
	 * seat_created - A new seat has appeared
	 * @name: (none)
	 * @capabilities: (none)
	 *
	 * 
	 */
	void (*seat_created)(void *data,
			     struct ivi_input *ivi_input,
			     const char *name,
			     uint32_t capabilities);
	/**
	 * seat_capabilities - A seat's capabilities have changed
	 * @name: (none)
	 * @capabilities: (none)
	 *
	 * 
	 */
	void (*seat_capabilities)(void *data,
				  struct ivi_input *ivi_input,
				  const char *name,
				  uint32_t capabilities);
	/**
	 * seat_destroyed - A seat is no longer available
	 * @name: (none)
	 *
	 * 
	 */
	void (*seat_destroyed)(void *data,
			       struct ivi_input *ivi_input,
			       const char *name);
	/**
	 * input_focus - input focus of surface in ivi compositor has
	 *	changed
	 * @surface: (none)
	 * @device: (none)
	 * @enabled: (none)
	 *
	 * The new input focus state is provided in argument enabled: If
	 * enabled is ILM_TRUE, this surface now has input focus enabled.
	 * If enabled is not ILM_TRUE, this surface no longer has input
	 * focus.
	 */
	void (*input_focus)(void *data,
			    struct ivi_input *ivi_input,
			    uint32_t surface,
			    uint32_t device,
			    int32_t enabled);
	/**
	 * input_acceptance - input acceptance has changed
	 * @surface: (none)
	 * @seat: (none)
	 * @accepted: (none)
	 *
	 * A surface has changed its input acceptance for a specific
	 * seat. If argument 'accepted' is ILM_TRUE, the surface now
	 * accepts the seat. If argument 'accepted' is not ILM_TRUE, the
	 * surface no longer accepts the seat.
	 */
	void (*input_acceptance)(void *data,
				 struct ivi_input *ivi_input,
				 uint32_t surface,
				 const char *seat,
				 int32_t accepted);
};

static inline int
ivi_input_add_listener(struct ivi_input *ivi_input,
		       const struct ivi_input_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) ivi_input,
				     (void (**)(void)) listener, data);
}

#define IVI_INPUT_SET_INPUT_FOCUS	0
#define IVI_INPUT_SET_INPUT_ACCEPTANCE	1

static inline void
ivi_input_set_user_data(struct ivi_input *ivi_input, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) ivi_input, user_data);
}

static inline void *
ivi_input_get_user_data(struct ivi_input *ivi_input)
{
	return wl_proxy_get_user_data((struct wl_proxy *) ivi_input);
}

static inline void
ivi_input_destroy(struct ivi_input *ivi_input)
{
	wl_proxy_destroy((struct wl_proxy *) ivi_input);
}

static inline void
ivi_input_set_input_focus(struct ivi_input *ivi_input, uint32_t surface, uint32_t device, int32_t enabled)
{
	wl_proxy_marshal((struct wl_proxy *) ivi_input,
			 IVI_INPUT_SET_INPUT_FOCUS, surface, device, enabled);
}

static inline void
ivi_input_set_input_acceptance(struct ivi_input *ivi_input, uint32_t surface, const char *seat, int32_t accepted)
{
	wl_proxy_marshal((struct wl_proxy *) ivi_input,
			 IVI_INPUT_SET_INPUT_ACCEPTANCE, surface, seat, accepted);
}

#ifdef  __cplusplus
}
#endif

#endif
